# SPDX-License-Identifier: BSD-2-Clause

__version__: str

class WASMInstance():
	def __init__(self) -> None:
		...

class WASMValue():
	def __init__(self, instance: WASMInstance, length: int, offset: int, value: int) -> None:
		...

	def set(self, value: int) -> None:
		...

	def get(self) -> int:
		...

class WASMRunner():
	def __init__(self, src: str, instance: WASMInstance, callback) -> None:
		...

	def __call__(self) -> int:
		...



# SPDX-License-Identifier: BSD-2-Clause

from torii.hdl.ast   import Cat, Const, Signal
from torii.hdl.cd    import ClockDomain
from torii.hdl.dsl   import Module

from ._harness_types import SimulatorRegressionTestMixinBase

class SimulatorRegressionTestMixin(SimulatorRegressionTestMixinBase):
	def test_bug_325(self):
		dut = Module()
		dut.d.comb += Signal().eq(Cat())
		self.get_simulator(dut).run()

	def test_bug_473(self):
		sim = self.get_simulator(Module())

		def process():
			self.assertEqual((yield -(Const(0b11, 2).as_signed())), 1)
		sim.add_process(process)
		sim.run()

	def test_bug_588(self):
		dut = Module()
		a = Signal(32)
		b = Signal(32)
		z = Signal(32)
		dut.d.comb += z.eq(a << b)
		with self.assertRaisesRegex(
			OverflowError,
			r'^Value defined at .+?[\\/]regression_harness\.py:\d+ is 4294967327 bits wide, '
			r'and wasm backend only supports signals that are less than 64 bits wide$'
		):
			self.get_simulator(dut)

	def test_bug_566(self):
		dut = Module()
		dut.d.sync += Signal().eq(0)
		sim = self.get_simulator(dut)
		with self.assertWarnsRegex(
			UserWarning,
			r'^Adding a clock process that drives a clock domain object named \'sync\', '
			r'which is distinct from an identically named domain in the simulated design$'
		):
			sim.add_clock(1e-6, domain = ClockDomain('sync'))

	def test_bug_826(self):
		sim = self.get_simulator(Module())

		def process():
			self.assertEqual((yield Const(0b0000, 4) | ~Const(1, 1)), 0b0000)
			self.assertEqual((yield Const(0b1111, 4) & ~Const(1, 1)), 0b0000)
			self.assertEqual((yield Const(0b1111, 4) ^ ~Const(1, 1)), 0b1111)

		sim.add_process(process)
		sim.run()



# SPDX-License-Identifier: BSD-2-Clause

import os
from warnings        import catch_warnings

from torii.hdl.ast   import Array, Fell, Past, Rose, Signal, Stable, ValueCastable,  unsigned
from torii.hdl.cd    import ClockDomain
from torii.hdl.dsl   import Module
from torii.hdl.mem   import Memory
from torii.sim       import Delay, Passive, Settle, Simulator, Tick

from ._harness_types import SimulatorIntegrationTestMixinBase

class SimulatorIntegrationTestsMixin(SimulatorIntegrationTestMixinBase):

	def setUp_counter(self):
		self.count = Signal(3, reset = 4)
		self.sync  = ClockDomain()

		self.m = Module()
		self.m.d.sync  += self.count.eq(self.count + 1)
		self.m.domains += self.sync

	def test_counter_process(self):
		self.setUp_counter()
		with self.assertSimulation(self.m) as sim:

			def process():
				self.assertEqual((yield self.count), 4)
				yield Delay(1e-6)
				self.assertEqual((yield self.count), 4)
				yield self.sync.clk.eq(1)
				self.assertEqual((yield self.count), 4)
				yield Settle()
				self.assertEqual((yield self.count), 5)
				yield Delay(1e-6)
				self.assertEqual((yield self.count), 5)
				yield self.sync.clk.eq(0)
				self.assertEqual((yield self.count), 5)
				yield Settle()
				self.assertEqual((yield self.count), 5)
				for _ in range(3):
					yield Delay(1e-6)
					yield self.sync.clk.eq(1)
					yield Delay(1e-6)
					yield self.sync.clk.eq(0)
				self.assertEqual((yield self.count), 0)

			sim.add_process(process)

	def test_counter_clock_and_sync_process(self):
		self.setUp_counter()
		with self.assertSimulation(self.m) as sim:
			sim.add_clock(1e-6, domain = 'sync')

			def process():
				self.assertEqual((yield self.count), 4)
				self.assertEqual((yield self.sync.clk), 1)
				yield
				self.assertEqual((yield self.count), 5)
				self.assertEqual((yield self.sync.clk), 1)
				for _ in range(3):
					yield
				self.assertEqual((yield self.count), 0)

			sim.add_sync_process(process)

	def test_reset(self):
		self.setUp_counter()
		sim = Simulator(self.m)
		sim.add_clock(1e-6)
		times = 0

		def process():
			nonlocal times
			self.assertEqual((yield self.count), 4)
			yield
			self.assertEqual((yield self.count), 5)
			yield
			self.assertEqual((yield self.count), 6)
			yield
			times += 1

		sim.add_sync_process(process)
		sim.run()
		sim.reset()
		sim.run()
		self.assertEqual(times, 2)

	def setUp_alu(self):
		self.a = Signal(8)
		self.b = Signal(8)
		self.o = Signal(8)
		self.x = Signal(8)
		self.s = Signal(2)
		self.sync = ClockDomain(reset_less = True)

		self.m = Module()
		self.m.d.comb += self.x.eq(self.a ^ self.b)
		with self.m.Switch(self.s):
			with self.m.Case(0):
				self.m.d.sync += self.o.eq(self.a + self.b)
			with self.m.Case(1):
				self.m.d.sync += self.o.eq(self.a - self.b)
			with self.m.Default():
				self.m.d.sync += self.o.eq(0)
		self.m.domains += self.sync

	def test_alu(self):
		self.setUp_alu()
		with self.assertSimulation(self.m) as sim:
			sim.add_clock(1e-6)

			def process():
				yield self.a.eq(5)
				yield self.b.eq(1)
				yield
				self.assertEqual((yield self.x), 4)
				yield
				self.assertEqual((yield self.o), 6)
				yield self.s.eq(1)
				yield
				yield
				self.assertEqual((yield self.o), 4)
				yield self.s.eq(2)
				yield
				yield
				self.assertEqual((yield self.o), 0)

			sim.add_sync_process(process)

	def setUp_clock_phase(self):
		self.m = Module()
		self.phase0 = self.m.domains.phase0 = ClockDomain()
		self.phase90 = self.m.domains.phase90 = ClockDomain()
		self.phase180 = self.m.domains.phase180 = ClockDomain()
		self.phase270 = self.m.domains.phase270 = ClockDomain()
		self.check = self.m.domains.check = ClockDomain()

		self.expected = [
			[0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0],
			[0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1],
			[0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1],
			[0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0],
		]

	def test_clock_phase(self):
		self.setUp_clock_phase()
		with self.assertSimulation(self.m) as sim:
			period = 1
			sim.add_clock(period / 8, phase = 0,              domain = 'check')
			sim.add_clock(period,     phase = 0 * period / 4, domain = 'phase0')
			sim.add_clock(period,     phase = 1 * period / 4, domain = 'phase90')
			sim.add_clock(period,     phase = 2 * period / 4, domain = 'phase180')
			sim.add_clock(period,     phase = 3 * period / 4, domain = 'phase270')

			def proc():
				clocks = [
					self.phase0.clk,
					self.phase90.clk,
					self.phase180.clk,
					self.phase270.clk
				]
				for i in range(16):
					yield
					for j, c in enumerate(clocks):
						self.assertEqual((yield c), self.expected[j][i])

			sim.add_sync_process(proc, domain = 'check')

	def setUp_multiclock(self):
		self.sys = ClockDomain()
		self.pix = ClockDomain()

		self.m = Module()
		self.m.domains += self.sys, self.pix

	def test_multiclock(self):
		self.setUp_multiclock()
		with self.assertSimulation(self.m) as sim:
			sim.add_clock(1e-6, domain = 'sys')
			sim.add_clock(0.3e-6, domain = 'pix')

			def sys_process():
				yield Passive()
				yield
				yield
				self.fail()

			def pix_process():
				yield
				yield
				yield

			sim.add_sync_process(sys_process, domain = 'sys')
			sim.add_sync_process(pix_process, domain = 'pix')

	def setUp_lhs_rhs(self):
		self.i = Signal(8)
		self.o = Signal(8)

		self.m = Module()
		self.m.d.comb += self.o.eq(self.i)

	def test_complex_lhs_rhs(self):
		self.setUp_lhs_rhs()
		with self.assertSimulation(self.m) as sim:
			def process():
				yield self.i.eq(0b10101010)
				yield self.i[:4].eq(-1)
				yield Settle()
				self.assertEqual((yield self.i[:4]), 0b1111)
				self.assertEqual((yield self.i), 0b10101111)
			sim.add_process(process)

	def test_run_until(self):
		m = Module()
		s = Signal()
		m.d.sync += s.eq(0)
		with self.assertSimulation(m, deadline = 100e-6) as sim:
			sim.add_clock(1e-6)

			def process():
				for _ in range(101):
					yield Delay(1e-6)
				self.fail()
			sim.add_process(process)

	def test_run_until_fail(self):
		m = Module()
		s = Signal()
		m.d.sync += s.eq(0)
		with self.assertRaises(AssertionError):
			with self.assertSimulation(m, deadline = 100e-6) as sim:
				sim.add_clock(1e-6)

				def process():
					for _ in range(99):
						yield Delay(1e-6)
					self.fail()
				sim.add_process(process)

	def test_add_process_wrong(self):
		with self.assertSimulation(Module()) as sim:
			with self.assertRaisesRegex(
				TypeError,
				r'^Cannot add a process 1 because it is not a generator function$'
			):
				sim.add_process(1)

	def test_add_process_wrong_generator(self):
		with self.assertSimulation(Module()) as sim:
			with self.assertRaisesRegex(
				TypeError,
				r'^Cannot add a process <.+?> because it is not a generator function$'
			):
				def process():
					yield Delay()
				sim.add_process(process())

	def test_add_clock_wrong_twice(self):
		m = Module()
		s = Signal()
		m.d.sync += s.eq(0)
		with self.assertSimulation(m) as sim:
			sim.add_clock(1)
			with self.assertRaisesRegex(
				ValueError,
				r'^Domain \'sync\' already has a clock driving it$'
			):
				sim.add_clock(1)

	def test_add_clock_wrong_missing(self):
		m = Module()
		with self.assertSimulation(m) as sim:
			with self.assertRaisesRegex(
				ValueError,
				r'^Domain \'sync\' is not present in simulation$'
			):
				sim.add_clock(1)

	def test_add_clock_if_exists(self):
		m = Module()
		with self.assertSimulation(m) as sim:
			sim.add_clock(1, if_exists = True)

	def test_command_wrong(self):
		survived = False
		with self.assertSimulation(Module()) as sim:
			def process():
				nonlocal survived

				with self.assertRaisesRegex(
					TypeError,
					r'Received unsupported command 1 from process .+'
				):
					yield 1
				# yield Settle()
				survived = True
			sim.add_process(process)
		self.assertTrue(survived)

	def test_value_castable(self):
		class MyValue(ValueCastable):
			@ValueCastable.lowermethod
			def as_value(self):
				return Signal()

			def shape():
				return unsigned(1)

		a = Array([1, 2, 3])
		a[MyValue()]

		survived = False
		with self.assertSimulation(Module()) as sim:
			def process():
				nonlocal survived
				yield MyValue()
				survived = True
			sim.add_process(process)
		self.assertTrue(survived)

	def setUp_memory(self, rd_synchronous = True, rd_transparent = True, wr_granularity = None):
		self.m = Module()
		self.memory = Memory(width = 8, depth = 4, init = [0xaa, 0x55])
		self.m.submodules.rdport = self.rdport = self.memory.read_port(
			domain = 'sync' if rd_synchronous else 'comb',
			transparent = rd_transparent
		)
		self.m.submodules.wrport = self.wrport = self.memory.write_port(
			granularity = wr_granularity
		)

	def test_memory_init(self):
		self.setUp_memory()
		with self.assertSimulation(self.m) as sim:
			def process():
				yield self.rdport.addr.eq(1)
				yield
				yield
				self.assertEqual((yield self.rdport.data), 0x55)
				yield self.rdport.addr.eq(2)
				yield
				yield
				self.assertEqual((yield self.rdport.data), 0x00)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_memory_write(self):
		self.setUp_memory()
		with self.assertSimulation(self.m) as sim:
			def process():
				yield self.wrport.addr.eq(4)
				yield self.wrport.data.eq(0x33)
				yield self.wrport.en.eq(1)
				yield
				yield self.wrport.en.eq(0)
				yield self.rdport.addr.eq(4)
				yield
				self.assertEqual((yield self.rdport.data), 0x33)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_memory_write_granularity(self):
		self.setUp_memory(wr_granularity = 4)
		with self.assertSimulation(self.m) as sim:
			def process():
				yield self.wrport.data.eq(0x50)
				yield self.wrport.en.eq(0b00)
				yield
				yield self.wrport.en.eq(0)
				yield
				self.assertEqual((yield self.rdport.data), 0xaa)
				yield self.wrport.en.eq(0b10)
				yield
				yield self.wrport.en.eq(0)
				yield
				self.assertEqual((yield self.rdport.data), 0x5a)
				yield self.wrport.data.eq(0x33)
				yield self.wrport.en.eq(0b01)
				yield
				yield self.wrport.en.eq(0)
				yield
				self.assertEqual((yield self.rdport.data), 0x53)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_memory_read_before_write(self):
		self.setUp_memory(rd_transparent = False)
		with self.assertSimulation(self.m) as sim:
			def process():
				yield self.wrport.data.eq(0x33)
				yield self.wrport.en.eq(1)
				yield
				self.assertEqual((yield self.rdport.data), 0xaa)
				yield
				self.assertEqual((yield self.rdport.data), 0xaa)
				yield Settle()
				self.assertEqual((yield self.rdport.data), 0x33)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_memory_write_through(self):
		self.setUp_memory(rd_transparent = True)
		with self.assertSimulation(self.m) as sim:
			def process():
				yield self.wrport.data.eq(0x33)
				yield self.wrport.en.eq(1)
				yield
				self.assertEqual((yield self.rdport.data), 0xaa)
				yield Settle()
				self.assertEqual((yield self.rdport.data), 0x33)
				yield
				yield self.rdport.addr.eq(1)
				yield Settle()
				self.assertEqual((yield self.rdport.data), 0x33)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_memory_async_read_write(self):
		self.setUp_memory(rd_synchronous = False)
		with self.assertSimulation(self.m) as sim:
			def process():
				yield self.rdport.addr.eq(0)
				yield Settle()
				self.assertEqual((yield self.rdport.data), 0xaa)
				yield self.rdport.addr.eq(1)
				yield Settle()
				self.assertEqual((yield self.rdport.data), 0x55)
				yield self.rdport.addr.eq(0)
				yield self.wrport.addr.eq(0)
				yield self.wrport.data.eq(0x33)
				yield self.wrport.en.eq(1)
				yield Tick('sync')
				self.assertEqual((yield self.rdport.data), 0xaa)
				yield Settle()
				self.assertEqual((yield self.rdport.data), 0x33)
			sim.add_clock(1e-6)
			sim.add_process(process)

	def test_memory_read_only(self):
		self.m = Module()
		self.memory = Memory(width = 8, depth = 4, init = [0xaa, 0x55])
		self.m.submodules.rdport = self.rdport = self.memory.read_port()
		with self.assertSimulation(self.m) as sim:
			def process():
				yield
				self.assertEqual((yield self.rdport.data), 0xaa)
				yield self.rdport.addr.eq(1)
				yield
				yield
				self.assertEqual((yield self.rdport.data), 0x55)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_memory_transparency_simple(self):
		m = Module()
		init = (0x11, 0x22, 0x33, 0x44)
		m.submodules.memory = memory = Memory(width = 8, depth = 4, init = init)
		rdport = memory.read_port()
		wrport = memory.write_port(granularity = 8)
		with self.assertSimulation(m) as sim:
			def process():
				yield rdport.addr.eq(0)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x11)
				yield rdport.addr.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x22)
				yield wrport.addr.eq(0)
				yield wrport.data.eq(0x44444444)
				yield wrport.en.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x22)
				yield wrport.addr.eq(1)
				yield wrport.data.eq(0x55)
				yield wrport.en.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x55)
				yield wrport.addr.eq(1)
				yield wrport.data.eq(0x66)
				yield wrport.en.eq(1)
				yield rdport.en.eq(0)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x55)
				yield wrport.addr.eq(2)
				yield wrport.data.eq(0x77)
				yield wrport.en.eq(1)
				yield rdport.en.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x66)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_memory_transparency_multibit(self):
		m = Module()
		init = (0x11111111, 0x22222222, 0x33333333, 0x44444444)
		m.submodules.memory = memory = Memory(width = 32, depth = 4, init = init)
		rdport = memory.read_port()
		wrport = memory.write_port(granularity = 8)
		with self.assertSimulation(m) as sim:
			def process():
				yield rdport.addr.eq(0)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x11111111)
				yield rdport.addr.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x22222222)
				yield wrport.addr.eq(0)
				yield wrport.data.eq(0x44444444)
				yield wrport.en.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x22222222)
				yield wrport.addr.eq(1)
				yield wrport.data.eq(0x55555555)
				yield wrport.en.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x22222255)
				yield wrport.addr.eq(1)
				yield wrport.data.eq(0x66666666)
				yield wrport.en.eq(2)
				yield rdport.en.eq(0)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x22222255)
				yield wrport.addr.eq(1)
				yield wrport.data.eq(0x77777777)
				yield wrport.en.eq(4)
				yield rdport.en.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x22776655)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_sample_helpers(self):
		m = Module()
		s = Signal(2)

		def mk(x):
			y = Signal.like(x)
			m.d.comb += y.eq(x)
			return y

		p0, r0, f0, s0 = mk(Past(s, 0)), mk(Rose(s)),    mk(Fell(s)),    mk(Stable(s))
		p1, r1, f1, s1 = mk(Past(s)),    mk(Rose(s, 1)), mk(Fell(s, 1)), mk(Stable(s, 1))
		p2, r2, f2, s2 = mk(Past(s, 2)), mk(Rose(s, 2)), mk(Fell(s, 2)), mk(Stable(s, 2))
		p3, r3, f3, s3 = mk(Past(s, 3)), mk(Rose(s, 3)), mk(Fell(s, 3)), mk(Stable(s, 3))
		with self.assertSimulation(m) as sim:
			def process_gen():
				yield s.eq(0b10)
				yield
				yield
				yield s.eq(0b01)
				yield

			def process_check():
				yield
				yield
				yield

				self.assertEqual((yield p0), 0b01)
				self.assertEqual((yield p1), 0b10)
				self.assertEqual((yield p2), 0b10)
				self.assertEqual((yield p3), 0b00)

				self.assertEqual((yield s0), 0b0)
				self.assertEqual((yield s1), 0b1)
				self.assertEqual((yield s2), 0b0)
				self.assertEqual((yield s3), 0b1)

				self.assertEqual((yield r0), 0b01)
				self.assertEqual((yield r1), 0b00)
				self.assertEqual((yield r2), 0b10)
				self.assertEqual((yield r3), 0b00)

				self.assertEqual((yield f0), 0b10)
				self.assertEqual((yield f1), 0b00)
				self.assertEqual((yield f2), 0b00)
				self.assertEqual((yield f3), 0b00)
			sim.add_clock(1e-6)
			sim.add_sync_process(process_gen)
			sim.add_sync_process(process_check)

	def test_vcd_wrong_nonzero_time(self):
		s = Signal()
		m = Module()
		m.d.sync += s.eq(s)
		sim = Simulator(m)
		sim.add_clock(1e-6)
		sim.run_until(1e-5)
		with self.assertRaisesRegex(
			ValueError,
			r'^Cannot start writing waveforms after advancing simulation time$'
		):
			with open(os.path.devnull, 'w') as f:
				with sim.write_vcd(f):
					pass # :nocov:

	def test_no_negated_boolean_warning(self):
		m = Module()
		a = Signal()
		b = Signal()
		m.d.comb += a.eq(~(b == b))
		with catch_warnings(record=True) as warns:
			Simulator(m).run()
			self.assertEqual(warns, [])

	def test_large_expr_parser_overflow(self):
		m = Module()
		a = Signal()

		op = a
		for _ in range(50):
			op = (op ^ 1)

		op = op & op

		m.d.comb += a.eq(op)
		Simulator(m)

	def test_switch_zero(self):
		m = Module()
		a = Signal(0)
		o = Signal()
		with m.Switch(a):
			with m.Case(''):
				m.d.comb += o.eq(1)
		with self.assertSimulation(m) as sim:
			def process():
				yield Settle()
				self.assertEqual((yield o), 1)
			sim.add_process(process)

	def test_switch_any(self):
		m = Module()
		a = Signal(3)
		o = Signal(3)
		with m.Switch(a):
			with m.Case('0-0'):
				m.d.comb += o.eq(0b000)
			with m.Case('0-1'):
				m.d.comb += o.eq(0b001)
			with m.Case('1-0'):
				m.d.comb += o.eq(0b100)
			with m.Case('1-1'):
				m.d.comb += o.eq(0b101)

		with self.assertSimulation(m) as sim:
			def process():
				yield a.eq(0b000)
				yield Settle()
				self.assertEqual((yield o), 0)
				yield a.eq(0b010)
				yield Settle()
				self.assertEqual((yield o), 0)
				yield a.eq(0b001)
				yield Settle()
				self.assertEqual((yield o), 1)
				yield a.eq(0b011)
				yield Settle()
				self.assertEqual((yield o), 1)
				yield a.eq(0b100)
				yield Settle()
				self.assertEqual((yield o), 0b100)
				yield a.eq(0b110)
				yield Settle()
				self.assertEqual((yield o), 0b100)
				yield a.eq(0b101)
				yield Settle()
				self.assertEqual((yield o), 0b101)
				yield a.eq(0b111)
				yield Settle()
				self.assertEqual((yield o), 0b101)

			sim.add_process(process)



# SPDX-License-Identifier: BSD-2-Clause

from torii.hdl.ast   import Array, Cat, Const, Mux, Signal, signed, unsigned
from torii.hdl.rec   import Record

from ._harness_types import SimulatorUnitTestMixinBase

class SimulatorUnitTestsMixin(SimulatorUnitTestMixinBase):

	def test_invert(self):
		self.assertStatement(
			lambda y, a: y.eq(~a),
			[Const(0b0000, 4)], Const(0b1111, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(~a),
			[Const(0b1010, 4)], Const(0b0101, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(~a),
			[Const(0, 4)], Const(-1, 4)
		)

	def test_neg(self):
		self.assertStatement(
			lambda y, a: y.eq(-a),
			[Const(0b0000, 4)], Const(0b0000, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(-a),
			[Const(0b0001, 4)], Const(0b1111, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(-a),
			[Const(0b1010, 4)], Const(0b0110, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(-a),
			[Const(1, 4)], Const(-1, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(-a),
			[Const(5, 4)], Const(-5, 4)
		)

	def test_bool(self):
		self.assertStatement(
			lambda y, a: y.eq(a.bool()),
			[Const(0, 4)], Const(0)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.bool()),
			[Const(1, 4)], Const(1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.bool()),
			[Const(2, 4)], Const(1)
		)

	def test_as_unsigned(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a.as_unsigned() == b),
			[Const(0b01, signed(2)), Const(0b0001, unsigned(4))], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a.as_unsigned() == b),
			[Const(0b11, signed(2)), Const(0b0011, unsigned(4))], Const(1)
		)

	def test_as_unsigned_lhs(self):
		self.assertStatement(
			lambda y, a: y.as_unsigned().eq(a),
			[Const(0b01, unsigned(2))], Const(0b0001, signed(4))
		)

	def test_as_signed(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a.as_signed() == b),
			[Const(0b01, unsigned(2)), Const(0b0001, signed(4))], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a.as_signed() == b),
			[Const(0b11, unsigned(2)), Const(0b1111, signed(4))], Const(1)
		)

	def test_as_signed_issue_502(self):
		self.assertStatement(
			lambda y, a: y.eq(a.as_signed()),
			[Const(0b01, unsigned(2))], Const(0b0001, signed(4))
		)
		self.assertStatement(
			lambda y, a: y.eq(a.as_signed()),
			[Const(0b11, unsigned(2))], Const(0b1111, signed(4))
		)

	def test_as_signed_lhs(self):
		self.assertStatement(
			lambda y, a: y.as_signed().eq(a),
			[Const(0b01, unsigned(2))], Const(0b0001, signed(4))
		)

	def test_any(self):
		self.assertStatement(
			lambda y, a: y.eq(a.any()),
			[Const(0b00, 2)], Const(0)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.any()),
			[Const(0b01, 2)], Const(1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.any()),
			[Const(0b10, 2)], Const(1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.any()),
			[Const(0b11, 2)], Const(1)
		)

	def test_all(self):
		self.assertStatement(
			lambda y, a: y.eq(a.all()),
			[Const(0b00, 2)], Const(0)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.all()),
			[Const(0b01, 2)], Const(0)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.all()),
			[Const(0b10, 2)], Const(0)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.all()),
			[Const(0b11, 2)], Const(1)
		)

	def test_xor_unary(self):
		self.assertStatement(
			lambda y, a: y.eq(a.xor()),
			[Const(0b00, 2)], Const(0)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.xor()),
			[Const(0b01, 2)], Const(1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.xor()),
			[Const(0b10, 2)], Const(1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.xor()),
			[Const(0b11, 2)], Const(0)
		)

	def test_add(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a + b),
			[Const(0,  4), Const(1, 4)], Const(1,   4)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a + b),
			[Const(-5, 4), Const(-5, 4)], Const(-10, 5)
		)

	def test_sub(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a - b),
			[Const(2, 4), Const(1, 4)], Const(1,   4)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a - b),
			[Const(0, 4), Const(1, 4)], Const(-1,  4)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a - b),
			[Const(0, 4), Const(10, 4)], Const(-10, 5)
		)

	def test_mul(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a * b),
			[Const(2, 4), Const(1, 4)], Const(2, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a * b),
			[Const(2, 4), Const(2, 4)], Const(4, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a * b),
			[Const(7, 4), Const(7, 4)], Const(49, 8)
		)

	def test_floordiv(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a // b),
			[Const(2, 4), Const(1, 4)], Const(2, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a // b),
			[Const(2, 4), Const(2, 4)], Const(1, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a // b),
			[Const(7, 4), Const(2, 4)], Const(3, 8)
		)

	def test_floordiv_neg(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a // b),
			[Const(-5, 4), Const(2, 4)], Const(-3, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a // b),
			[Const(-5, 4), Const(-2, 4)], Const(2, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a // b),
			[Const(5, 4), Const(2, 4)], Const(2, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a // b),
			[Const(5, 4), Const(-2, 4)], Const(-3, 8)
		)

	def test_mod(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(2, 4), Const(0, 4)], Const(0, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(2, 4), Const(1, 4)], Const(0, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(2, 4), Const(2, 4)], Const(0, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(7, 4), Const(2, 4)], Const(1, 8)
		)

	def test_mod_neg(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(-5, 4), Const(3, 4)], Const(1, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(-5, 4), Const(-3, 4)], Const(-2, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(5, 4), Const(3, 4)], Const(2, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(5, 4), Const(-3, 4)], Const(-1, 8)
		)

	def test_and(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a & b),
			[Const(0b1100, 4), Const(0b1010, 4)], Const(0b1000, 4)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a & b),
			[Const(0b1010, 4), Const(0b10, signed(2))], Const(0b1010, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(a),
			[Const(0b1010, 4) & Const(-2, 2).as_unsigned()], Const(0b0010, 4)
		)

	def test_or(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a | b),
			[Const(0b1100, 4), Const(0b1010, 4)], Const(0b1110, 4)
		)

	def test_xor_binary(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a ^ b),
			[Const(0b1100, 4), Const(0b1010, 4)], Const(0b0110, 4)
		)

	def test_shl(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a << b),
			[Const(0b1001, 4), Const(0)],  Const(0b1001, 5)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a << b),
			[Const(0b1001, 4), Const(3)],  Const(0b1001000, 7)
		)

	def test_shr(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a >> b),
			[Const(0b1001, 4), Const(0)],  Const(0b1001, 4)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a >> b),
			[Const(0b1001, 4), Const(2)],  Const(0b10,  4)
		)

	def test_eq(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a == b),
			[Const(0, 4), Const(0, 4)], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a == b),
			[Const(0, 4), Const(1, 4)], Const(0)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a == b),
			[Const(1, 4), Const(0, 4)], Const(0)
		)

	def test_ne(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a != b),
			[Const(0, 4), Const(0, 4)], Const(0)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a != b),
			[Const(0, 4), Const(1, 4)], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a != b),
			[Const(1, 4), Const(0, 4)], Const(1)
		)

	def test_lt(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a < b),
			[Const(0, 4), Const(0, 4)], Const(0)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a < b),
			[Const(0, 4), Const(1, 4)], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a < b),
			[Const(1, 4), Const(0, 4)], Const(0)
		)

	def test_ge(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a >= b),
			[Const(0, 4), Const(0, 4)], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a >= b),
			[Const(0, 4), Const(1, 4)], Const(0)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a >= b),
			[Const(1, 4), Const(0, 4)], Const(1)
		)

	def test_gt(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a > b),
			[Const(0, 4), Const(0, 4)], Const(0)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a > b),
			[Const(0, 4), Const(1, 4)], Const(0)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a > b),
			[Const(1, 4), Const(0, 4)], Const(1)
		)

	def test_le(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a <= b),
			[Const(0, 4), Const(0, 4)], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a <= b),
			[Const(0, 4), Const(1, 4)], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a <= b),
			[Const(1, 4), Const(0, 4)], Const(0)
		)

	def test_mux(self):
		self.assertStatement(
			lambda y, a, b, c: y.eq(Mux(c, a, b)),
			[Const(2, 4), Const(3, 4), Const(0)], Const(3, 4)
		)
		self.assertStatement(
			lambda y, a, b, c: y.eq(Mux(c, a, b)),
			[Const(2, 4), Const(3, 4), Const(1)], Const(2, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(a),
			[Mux(0, Const(0b1010, 4), Const(0b10, 2).as_signed())], Const(0b1110, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(a),
			[Mux(0, Const(0b1010, 4), Const(-2, 2).as_unsigned())], Const(0b0010, 4)
		)

	def test_mux_invert(self):
		self.assertStatement(
			lambda y, a, b, c: y.eq(Mux(~c, a, b)),
			[Const(2, 4), Const(3, 4), Const(0)], Const(2, 4)
		)
		self.assertStatement(
			lambda y, a, b, c: y.eq(Mux(~c, a, b)),
			[Const(2, 4), Const(3, 4), Const(1)], Const(3, 4)
		)

	def test_mux_wide(self):
		self.assertStatement(
			lambda y, a, b, c: y.eq(Mux(c, a, b)),
			[Const(2, 4), Const(3, 4), Const(0, 2)], Const(3, 4)
		)
		self.assertStatement(
			lambda y, a, b, c: y.eq(Mux(c, a, b)),
			[Const(2, 4), Const(3, 4), Const(1, 2)], Const(2, 4)
		)
		self.assertStatement(
			lambda y, a, b, c: y.eq(Mux(c, a, b)),
			[Const(2, 4), Const(3, 4), Const(2, 2)], Const(2, 4)
		)

	def test_abs(self):
		self.assertStatement(
			lambda y, a: y.eq(abs(a)),
			[Const(3,  unsigned(8))], Const(3, unsigned(8))
		)
		self.assertStatement(
			lambda y, a: y.eq(abs(a)),
			[Const(-3, unsigned(8))], Const(-3, unsigned(8))
		)
		self.assertStatement(
			lambda y, a: y.eq(abs(a)),
			[Const(3,  signed(8))], Const(3, signed(8))
		)
		self.assertStatement(
			lambda y, a: y.eq(abs(a)),
			[Const(-3, signed(8))], Const(3, signed(8))
		)

	def test_slice(self):
		self.assertStatement(
			lambda y, a: y.eq(a[2]),
			[Const(0b10110100, 8)], Const(0b1,  1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a[2:4]),
			[Const(0b10110100, 8)], Const(0b01, 2)
		)

	def test_slice_lhs(self):
		self.assertStatement(
			lambda y, a: y[2].eq(a),
			[Const(0b0,  1)], Const(0b11111011, 8),
			reset = 0b11111111
		)
		self.assertStatement(
			lambda y, a: y[2:4].eq(a),
			[Const(0b01, 2)], Const(0b11110111, 8),
			reset = 0b11111011
		)

	def test_bit_select(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a.bit_select(b, 3)),
			[Const(0b10110100, 8), Const(0)], Const(0b100, 3)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a.bit_select(b, 3)),
			[Const(0b10110100, 8), Const(2)], Const(0b101, 3)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a.bit_select(b, 3)),
			[Const(0b10110100, 8), Const(3)], Const(0b110, 3)
		)

	def test_bit_select_lhs(self):
		self.assertStatement(
			lambda y, a, b: y.bit_select(a, 3).eq(b),
			[Const(0), Const(0b100, 3)], Const(0b11111100, 8),
			reset = 0b11111111
		)
		self.assertStatement(
			lambda y, a, b: y.bit_select(a, 3).eq(b),
			[Const(2), Const(0b101, 3)], Const(0b11110111, 8),
			reset = 0b11111111
		)
		self.assertStatement(
			lambda y, a, b: y.bit_select(a, 3).eq(b),
			[Const(3), Const(0b110, 3)], Const(0b11110111, 8),
			reset = 0b11111111
		)

	def test_word_select(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a.word_select(b, 3)),
			[Const(0b10110100, 8), Const(0)], Const(0b100, 3)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a.word_select(b, 3)),
			[Const(0b10110100, 8), Const(1)], Const(0b110, 3)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a.word_select(b, 3)),
			[Const(0b10110100, 8), Const(2)], Const(0b010, 3)
		)

	def test_word_select_lhs(self):
		self.assertStatement(
			lambda y, a, b: y.word_select(a, 3).eq(b),
			[Const(0), Const(0b100, 3)], Const(0b11111100, 8), reset = 0b11111111
		)
		self.assertStatement(
			lambda y, a, b: y.word_select(a, 3).eq(b),
			[Const(1), Const(0b101, 3)], Const(0b11101111, 8), reset = 0b11111111
		)
		self.assertStatement(
			lambda y, a, b: y.word_select(a, 3).eq(b),
			[Const(2), Const(0b110, 3)], Const(0b10111111, 8), reset = 0b11111111
		)

	def test_cat(self):
		self.assertStatement(
			lambda y, *xs: y.eq(Cat(*xs)),
			[Const(0b10, 2), Const(0b01, 2)], Const(0b0110, 4)
		)

	def test_cat_lhs(self):
		l = Signal(3) # noqa: E741
		m = Signal(3) # noqa: E741
		n = Signal(3) # noqa: E741
		self.assertStatement(
			lambda y, a: [Cat(l, m, n).eq(a), y.eq(Cat(n, m, l))],
			[Const(0b100101110, 9)], Const(0b110101100, 9)
		)

	def test_nested_cat_lhs(self):
		l = Signal(3) # noqa: E741
		m = Signal(3) # noqa: E741
		n = Signal(3) # noqa: E741
		self.assertStatement(
			lambda y, a: [Cat(Cat(l, Cat(m)), n).eq(a), y.eq(Cat(n, m, l))],
			[Const(0b100101110, 9)], Const(0b110101100, 9)
		)

	def test_record(self):
		rec = Record([
			('l', 1),
			('m', 2),
		])
		self.assertStatement(
			lambda y, a: [rec.eq(a), y.eq(rec)],
			[Const(0b101, 3)], Const(0b101, 3)
		)

	def test_replicate(self):
		self.assertStatement(
			lambda y, a: y.eq(a.replicate(3)),
			[Const(0b10, 2)], Const(0b101010, 6)
		)

	def test_array(self):
		array = Array([1, 4, 10])
		self.assertStatement(
			lambda y, a: y.eq(array[a]),
			[Const(0)], Const(1)
		)
		self.assertStatement(
			lambda y, a: y.eq(array[a]),
			[Const(1)], Const(4)
		)
		self.assertStatement(
			lambda y, a: y.eq(array[a]),
			[Const(2)], Const(10)
		)

	def test_array_oob(self):
		array = Array([1, 4, 10])
		self.assertStatement(
			lambda y, a: y.eq(array[a]),
			[Const(3)], Const(10)
		)
		self.assertStatement(
			lambda y, a: y.eq(array[a]),
			[Const(4)], Const(10)
		)

	def test_array_lhs(self):
		l = Signal(3, reset = 1) # noqa: E741
		m = Signal(3, reset = 4) # noqa: E741
		n = Signal(3, reset = 7) # noqa: E741

		array = Array([l, m, n])
		self.assertStatement(
			lambda y, a, b: [array[a].eq(b), y.eq(Cat(*array))],
			[Const(0), Const(0b000)], Const(0b111100000)
		)
		self.assertStatement(
			lambda y, a, b: [array[a].eq(b), y.eq(Cat(*array))],
			[Const(1), Const(0b010)], Const(0b111010001)
		)
		self.assertStatement(
			lambda y, a, b: [array[a].eq(b), y.eq(Cat(*array))],
			[Const(2), Const(0b100)], Const(0b100100001)
		)

	def test_array_lhs_oob(self):
		l = Signal(3) # noqa: E741
		m = Signal(3) # noqa: E741
		n = Signal(3) # noqa: E741
		array = Array([l, m, n])
		self.assertStatement(
			lambda y, a, b: [array[a].eq(b), y.eq(Cat(*array))],
			[Const(3), Const(0b001)], Const(0b001000000)
		)
		self.assertStatement(
			lambda y, a, b: [array[a].eq(b), y.eq(Cat(*array))],
			[Const(4), Const(0b010)], Const(0b010000000)
		)

	def test_array_index(self):
		array = Array(Array(x * y for y in range(10)) for x in range(10))
		for x in range(10):
			for y in range(10):
				self.assertStatement(
					lambda y, a, b: y.eq(array[a][b]),
					[Const(x), Const(y)], Const(x * y)
				)

	def test_array_attr(self):
		from collections import namedtuple
		pair = namedtuple('pair', ('p', 'n'))

		array = Array(pair(x, -x) for x in range(10))
		for i in range(10):
			self.assertStatement(
				lambda y, a: y.eq(array[a].p + array[a].n),
				[Const(i)], Const(0)
			)

	def test_shift_left(self):
		self.assertStatement(
			lambda y, a: y.eq(a.shift_left(1)),
			[Const(0b10100010, 8)], Const(   0b101000100, 9)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.shift_left(4)),
			[Const(0b10100010, 8)], Const(0b101000100000, 12)
		)

	def test_shift_right(self):
		self.assertStatement(
			lambda y, a: y.eq(a.shift_right(1)),
			[Const(0b10100010, 8)], Const(0b1010001, 7)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.shift_right(4)),
			[Const(0b10100010, 8)], Const(   0b1010, 4)
		)

	def test_rotate_left(self):
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(1)),
			[Const(0b1)], Const(0b1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(1)),
			[Const(0b1001000)], Const(0b0010001)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(5)),
			[Const(0b1000000)], Const(0b0010000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(5)),
			[Const(0b1000001)], Const(0b0110000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(7)),
			[Const(0b1000000)], Const(0b1000000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(7)),
			[Const(0b1000001)], Const(0b1000001)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(9)),
			[Const(0b1000000)], Const(0b0000010)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(9)),
			[Const(0b1000001)], Const(0b0000110)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-1)),
			[Const(0b1)], Const(0b1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-1)),
			[Const(0b1001000)], Const(0b0100100)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-5)),
			[Const(0b1000000)], Const(0b0000010)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-5)),
			[Const(0b1000001)], Const(0b0000110)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-7)),
			[Const(0b1000000)], Const(0b1000000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-7)),
			[Const(0b1000001)], Const(0b1000001)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-9)),
			[Const(0b1000000)], Const(0b0010000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-9)),
			[Const(0b1000001)], Const(0b0110000)
		)

	def test_rotate_right(self):

		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(1)),
			[Const(0b1)], Const(0b1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(1)),
			[Const(0b1001000)], Const(0b0100100)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(5)),
			[Const(0b1000000)], Const(0b0000010)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(5)),
			[Const(0b1000001)], Const(0b0000110)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(7)),
			[Const(0b1000000)], Const(0b1000000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(7)),
			[Const(0b1000001)], Const(0b1000001)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(9)),
			[Const(0b1000000)], Const(0b0010000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(9)),
			[Const(0b1000001)], Const(0b0110000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-1)),
			[Const(0b1)], Const(0b1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-1)),
			[Const(0b1001000)], Const(0b0010001)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-5)),
			[Const(0b1000000)], Const(0b0010000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-5)),
			[Const(0b1000001)], Const(0b0110000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-7)),
			[Const(0b1000000)], Const(0b1000000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-7)),
			[Const(0b1000001)], Const(0b1000001)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-9)),
			[Const(0b1000000)], Const(0b0000010)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-9)),
			[Const(0b1000001)], Const(0b0000110)
		)



# SPDX-License-Identifier: BSD-2-Clause

from collections.abc import Generator, Iterable
from contextlib      import contextmanager
from itertools       import chain
from re              import search
from typing          import IO

from vcd             import VCDWriter
from vcd.gtkw        import GTKWSave
from vcd.writer      import Variable
from torii.hdl.ast   import Signal, SignalDict, Value
from torii.hdl.ir    import Fragment
from torii.sim._base import BaseEngine, BaseSignalState, BaseSimulation

from ._wasm_engine   import __version__, WASMInstance, WASMValue
from .wasmrtl        import WASMFragmentCompiler
from .wasmclock      import WASMClockProcess
from .wasmcoro       import WASMCoroProcess

__all__ = (
	'WASMSimEngine',
)

__version__ = __version__

class _VCDWriter:
	@staticmethod
	def decode_to_vcd(signal, value):
		return signal.decoder(value).expandtabs().replace(' ', '_')

	def __init__(
		self, fragment: Fragment, *, vcd_file: IO | str | None, gtkw_file: IO | str | None = None,
		traces: Iterable[Signal] = ()
	) -> None:

		self._close_vcd = False
		self._close_gtkw = False
		if isinstance(vcd_file, str):
			vcd_file = open(vcd_file, 'w')
			self._close_vcd = True

		if isinstance(gtkw_file, str):
			gtkw_file = open(gtkw_file, 'w')
			self._close_gtkw = True

		self.vcd_vars = SignalDict[Variable]()
		self.vcd_file = vcd_file
		self.vcd_writer = None
		if vcd_file is not None:
			self.vcd_writer = VCDWriter(vcd_file, timescale = '1 ps', comment = 'Generated by Torii')

		self.gtkw_names = SignalDict[tuple[str, ...]]()
		self.gtkw_file = gtkw_file
		self.gtkw_save = None
		if gtkw_file is not None:
			self.gtkw_save = GTKWSave(gtkw_file)

		self.traces = []

		signal_names = SignalDict[set[tuple[str, ...]]]()
		for subfragment, subfragment_name in fragment._assign_names_to_fragments(
			hierarchy = ('bench', 'top', )
		).items():
			for signal, signal_name in subfragment._assign_names_to_signals().items():
				if signal not in signal_names:
					signal_names[signal] = set[tuple[str, ...]]()
				signal_names[signal].add((*subfragment_name, signal_name))

		trace_names = SignalDict[set[tuple[str, ...]]]()
		assigned_names = set[str]()
		for trace in traces:
			trace_value = Value.cast(trace)
			for trace_signal in trace_value._rhs_signals():
				assert trace_signal is not None and isinstance(trace_signal, Signal)
				if trace_signal not in signal_names:
					if trace_signal.name not in assigned_names:
						name = trace_signal.name
					else:
						name = f'{trace_signal.name}${len(assigned_names)}'
						if name in assigned_names:
							raise RuntimeError(
								f'Unable to create unique signal name for {trace_signal.name}. '
								f'Tried {name}, but it already exists.'
							)
					trace_names[trace_signal] = { ('bench', name) }
					assigned_names.add(name)
				self.traces.append(trace_signal)

		if self.vcd_writer is None:
			return

		for signal, names in chain(signal_names.items(), trace_names.items()):
			if signal.decoder:
				var_type = 'string'
				var_size = 1
				var_init = self.decode_to_vcd(signal, signal.reset)
			else:
				var_type = 'wire'
				var_size = signal.width
				var_init = signal.reset

			for (*var_scope, var_name) in names:
				if search(r'[ \t\r\n]', var_name):
					raise NameError(f'Signal \'{".".join(var_scope)}.{var_name}\' contains a whitespace character')

				suffix = None
				while True:
					try:
						if suffix is None:
							var_name_suffix = var_name
						else:
							var_name_suffix = f'{var_name}${suffix}'
						if signal not in self.vcd_vars:
							vcd_var = self.vcd_writer.register_var(
								scope = var_scope, name = var_name_suffix,
								var_type = var_type, size = var_size, init = var_init)
							self.vcd_vars[signal] = vcd_var
						else:
							self.vcd_writer.register_alias(
								scope = var_scope, name = var_name_suffix,
								var = self.vcd_vars[signal])
						break
					except KeyError:
						suffix = (suffix or 0) + 1

				if signal not in self.gtkw_names:
					self.gtkw_names[signal] = (*var_scope, var_name_suffix)

	def update(self, timestamp, signal, value):
		vcd_var = self.vcd_vars.get(signal)
		if vcd_var is None:
			return

		if signal.decoder:
			var_value = self.decode_to_vcd(signal, value)
		else:
			var_value = value
		self.vcd_writer.change(vcd_var, timestamp, var_value)

	def close(self, timestamp):
		if self.vcd_writer is not None:
			self.vcd_writer.close(timestamp)

		if self.gtkw_save is not None:
			self.gtkw_save.dumpfile(self.vcd_file.name)
			self.gtkw_save.dumpfile_size(self.vcd_file.tell())

			self.gtkw_save.treeopen('top')
			for signal in self.traces:
				if len(signal) > 1 and not signal.decoder:
					suffix = f'[{len(signal) - 1}:0]'
				else:
					suffix = ''
				self.gtkw_save.trace('.'.join(self.gtkw_names[signal]) + suffix)

		if self._close_vcd:
			self.vcd_file.close()
		if self._close_gtkw:
			self.gtkw_file.close()

class _Timeline:
	def __init__(self) -> None:
		self.now = 0
		self.deadlines = dict()

	def reset(self):
		self.now = 0
		self.deadlines.clear()

	def at(self, run_at, process):
		if process in self.deadlines:
			raise ValueError(f'Process {process!r} already in deadline list')
		self.deadlines[process] = run_at

	def delay(self, delay_by, process):
		if delay_by is None:
			run_at = self.now
		else:
			run_at = self.now + delay_by
		self.at(run_at, process)

	def advance(self):
		nearest_processes = set()
		nearest_deadline = None
		for process, deadline in self.deadlines.items():
			if deadline is None:
				if nearest_deadline is not None:
					nearest_processes.clear()
				nearest_processes.add(process)
				nearest_deadline = self.now
				break
			elif nearest_deadline is None or deadline <= nearest_deadline:
				if deadline < self.now:
					raise ValueError(f'Deadline of {deadline} is behind our current state of {self.now}')

				if nearest_deadline is not None and deadline < nearest_deadline:
					nearest_processes.clear()
				nearest_processes.add(process)
				nearest_deadline = deadline

		if not nearest_processes:
			return False

		for process in nearest_processes:
			process.runnable = True
			del self.deadlines[process]
		self.now = nearest_deadline

		return True

class _WASMGlobal():
	def __init__(self, memory: WASMInstance, signal, offset, value) -> None:
		self._memory = memory
		self._signal = signal
		self._wasm   = WASMValue(memory, len(self._signal), offset, value)
		self._value  = value
		self.set(value)

	def set(self, value):
		self._value = value & ((1 << len(self._signal)) - 1)
		self._wasm.set(value)

	def get(self):
		self._wasm.get()

	def value(self):
		return self._value

	def update(self, value):
		self._value = value

	def __eq__(self, other):
		if isinstance(other, _WASMGlobal):
			return self.value() == other.value()
		else:
			return self.value() == int(other)

class _WASMSignalState(BaseSignalState):
	__slots__ = ('signal', 'curr', 'next', 'waiters', 'pending')

	def __init__(self, memory: WASMInstance, index, signal, pending) -> None:
		self.signal = signal
		self.pending = pending
		self.waiters = dict()
		self.curr = _WASMGlobal(memory, signal, index * 2, signal.reset)
		self.next = _WASMGlobal(memory, signal, index * 2 + 1, signal.reset)

	def set(self, value):
		self.next.update(value)
		self.pending.add(self)

	def update(self, value):
		raw_val = int(value)
		if self.next.value() == raw_val:
			return

		self.next.set(raw_val)
		self.pending.add(self)

	def commit(self):
		if self.curr == self.next:
			return False
		self.curr.set(self.next.value())

		awoken_any = False
		for process, trigger in self.waiters.items():
			if trigger is None or trigger == self.curr.value():
				process.runnable = awoken_any = True
		return awoken_any

class _WASMimulation(BaseSimulation):
	def __init__(self) -> None:
		self.timeline = _Timeline()
		self.signals  = SignalDict()
		self.slots    = []
		self.pending  = set()
		self.memory = WASMInstance()

	def set_slot(self, index, value):
		self.slots[index].set(value)

	def get_signal(self, signal):
		try:
			return self.signals[signal]
		except KeyError:
			index = len(self.slots)
			self.slots.append(_WASMSignalState(self.memory, index, signal, self.pending))
			self.signals[signal] = index
			return index

	def commit(self, changed = None):
		converged = True
		for signal_state in self.pending:
			if signal_state.commit():
				converged = False
		if changed is not None:
			changed.update(self.pending)
		self.pending.clear()
		return converged

	def wait_interval(self, process, interval):
		self.timeline.delay(interval, process)

	def add_trigger(self, process, signal, *, trigger = None):
		index = self.get_signal(signal)
		if process in self.slots[index].waiters and self.slots[index].waiters[process] != trigger:
			raise ValueError('Unable to add trigger for process!')
		self.slots[index].waiters[process] = trigger

	def remove_trigger(self, process, signal):
		index = self.get_signal(signal)
		if process not in self.slots[index].waiters:
			raise ValueError(f'Unable to remove trigger for process {process!r}, not in the slot list')
		del self.slots[index].waiters[process]

class WASMSimEngine(BaseEngine):
	def __init__(self, fragment: Fragment) -> None:
		self._state = _WASMimulation()
		self._timeline = self._state.timeline
		self._frag = fragment
		self._processes = WASMFragmentCompiler(self._state)(self._frag)
		self._vcd_writers = []

	def add_coroutine_process(self, process, *, default_cmd):
		self._processes.add(
			WASMCoroProcess(self._state, self._frag.domains, process, default_cmd = default_cmd)
		)

	def add_clock_process(self, clock, *, phase, period):
		self._processes.add(
			WASMClockProcess(self._state, clock, phase = phase, period = period)
		)

	def reset(self):
		pass

	def _step(self):
		changed = set() if self._vcd_writers else None
		# Performs the two phases of a delta cycle in a loop:
		converged = False
		while not converged:
			# 1. eval: run and suspend every non-waiting process once, queueing signal changes
			for process in self._processes:
				if process.runnable:
					process.runnable = False
					process.run()

			# 2. commit: apply every queued signal change, waking up any waiting processes
			converged = self._state.commit(changed)

		for vcd_writer in self._vcd_writers:
			for signal_state in changed:
				vcd_writer.update(self._timeline.now, signal_state.signal, signal_state.curr.value())

	def advance(self):
		self._step()
		self._timeline.advance()
		return any(not process.passive for process in self._processes)

	@property
	def now(self):
		return self._timeline.now

	@contextmanager
	def write_vcd(
		self, *, vcd_file: IO | str | None, gtkw_file: IO | str | None = None,
		traces: Iterable[Signal]
	) -> Generator[None, None, None]:
		vcd_writer = _VCDWriter(self._frag, vcd_file = vcd_file, gtkw_file = gtkw_file, traces = traces)
		try:
			self._vcd_writers.append(vcd_writer)
			yield
		finally:
			vcd_writer.close(self._timeline.now)
			self._vcd_writers.remove(vcd_writer)



# SPDX-License-Identifier: BSD-2-Clause

from typing import TYPE_CHECKING

__all__ = (
	'SimulatorIntegrationTestMixinBase',
	'SimulatorRegressionTestMixinBase',
	'SimulatorUnitTestMixinBase',
)

# NOTE(aki):
# This is just applying typing stubs to ensure the type checker is happy, doesn't effect runtime
# and a chunk of this was taken from the typeshed stubs for `unittest.case.TestCase`.
if TYPE_CHECKING:
	from collections.abc import Callable
	from contextlib      import contextmanager
	from re              import Pattern
	from typing          import Any, Iterator, ParamSpec, TypeVar, overload
	from unittest.case   import _AssertRaisesContext, _AssertWarnsContext

	from torii.sim       import Simulator

	_E = TypeVar('_E', bound = BaseException)
	_P = ParamSpec("_P")

	class SimulatorIntegrationTestMixinBase:
		def fail(self) -> None:
			...

		@contextmanager
		def assertSimulation(self, module, deadline = None) -> Iterator[Simulator]:
			...

		def assertEqual(self, first: Any, second: Any, msg: Any = None) -> None:
			...

		def assertTrue(self, expr: Any, msg: Any = None) -> None:
			...

		@overload
		def assertRaises( # type: ignore
			self, expected_exception: type[BaseException] | tuple[type[BaseException], ...],
			callable: Callable[..., object], *args: Any, **kwargs: Any,
		) -> None:
			...

		@overload
		def assertRaises(
			self, expected_exception: type[_E] | tuple[type[_E], ...], *, msg: Any = ...
		) -> _AssertRaisesContext[_E]:
			...

		@overload
		def assertRaisesRegex( # type: ignore
			self, expected_exception: type[BaseException] | tuple[type[BaseException], ...],
			expected_regex: str | Pattern[str], callable: Callable[..., object], *args: Any, **kwargs: Any,
		) -> None:
			...

		@overload
		def assertRaisesRegex(
			self, expected_exception: type[_E] | tuple[type[_E], ...], expected_regex: str | Pattern[str], *,
			msg: Any = ...
		) -> _AssertRaisesContext[_E]:
			...

	class SimulatorRegressionTestMixinBase:
		def get_simulator(self, dut) -> Simulator:
			...

		def assertEqual(self, first: Any, second: Any, msg: Any = None) -> None:
			...

		@overload
		def assertRaisesRegex( # type: ignore
			self, expected_exception: type[BaseException] | tuple[type[BaseException], ...],
			expected_regex: str | Pattern[str], callable: Callable[..., object], *args: Any, **kwargs: Any,
		) -> None:
			...

		@overload
		def assertRaisesRegex(
			self, expected_exception: type[_E] | tuple[type[_E], ...], expected_regex: str | Pattern[str], *,
			msg: Any = ...
		) -> _AssertRaisesContext[_E]:
			...

		@overload
		def assertWarnsRegex( # type: ignore
			self, expected_warning: type[Warning] | tuple[type[Warning], ...],
			expected_regex: str | Pattern[str], callable: Callable[_P, object], *args: _P.args, **kwargs: _P.kwargs,
		) -> None:
			...

		@overload
		def assertWarnsRegex(
			self, expected_warning: type[Warning] | tuple[type[Warning], ...], expected_regex: str | Pattern[str], *,
			msg: Any = ...
		) -> _AssertWarnsContext:
			...

	class SimulatorUnitTestMixinBase:
		def assertStatement(self, stmt, inputs, output, reset = 0) -> None:
			...
else:
	SimulatorIntegrationTestMixinBase = object
	SimulatorRegressionTestMixinBase = object
	SimulatorUnitTestMixinBase = object



# SPDX-License-Identifier: BSD-2-Clause

from inspect         import getfile, getlineno, iscoroutine, isgenerator

from torii.hdl       import ClockDomain, Const, Value
from torii.hdl.ast   import SignalSet, Statement, ValueCastable
from torii.sim._base import BaseProcess
from torii.sim.core  import Active, Delay, Passive, Settle, Tick
from .wasmrtl        import _RHSValueCompiler, _StatementCompiler
from ._wasm_engine   import WASMRunner

__all__ = (
	'WASMCoroProcess',
)

def foo(one, two):
	assert False

class WASMCoroProcess(BaseProcess):
	def __init__(self, state, domains, constructor, *, default_cmd = None) -> None:
		self.state = state
		self.domains = domains
		self.constructor = constructor
		self.default_cmd = default_cmd

		self.reset()

	def reset(self):
		self.runnable = True
		self.passive = False

		self.coroutine = self.constructor()
		self.waits_on = SignalSet()

	def src_loc(self):
		coroutine = self.coroutine
		if coroutine is None:
			return None
		while coroutine.gi_yieldfrom is not None and isgenerator(coroutine.gi_yieldfrom):
			coroutine = coroutine.gi_yieldfrom
		if isgenerator(coroutine):
			frame = coroutine.gi_frame
		if iscoroutine(coroutine):
			frame = coroutine.cr_frame
		return f'{getfile(frame)}:{getlineno(frame)}'

	def add_trigger(self, signal, trigger = None):
		self.state.add_trigger(self, signal, trigger = trigger)
		self.waits_on.add(signal)

	def clear_triggers(self):
		for signal in self.waits_on:
			self.state.remove_trigger(self, signal)
		self.waits_on.clear()

	def run(self):
		if self.coroutine is None:
			return

		self.clear_triggers()

		response = None
		exception = None
		while True:
			try:
				if exception is None:
					command = self.coroutine.send(response)
				else:
					command = self.coroutine.throw(exception)
			except StopIteration:
				self.passive = True
				self.coroutine = None
				return

			try:
				if command is None:
					command = self.default_cmd
				response = None
				exception = None

				if isinstance(command, ValueCastable):
					command = Value.cast(command)
				if isinstance(command, Value):
					module_code = _RHSValueCompiler.compile(self.state, command, mode = 'curr')
					run = WASMRunner(module_code, self.state.memory, self.state.set_slot)
					result = run()
					response = Const.normalize(result, command.shape())

				elif isinstance(command, Statement):
					module_code = _StatementCompiler.compile(self.state, command)
					run = WASMRunner(module_code, self.state.memory, self.state.set_slot)
					run()

				elif type(command) is Tick:
					domain = command.domain
					if isinstance(domain, ClockDomain):
						pass
					elif domain in self.domains:
						domain = self.domains[domain]
					else:
						raise NameError(
							f'Received command {command!r} that refers to a nonexistent '
							f'domain {command.domain!r} from process {self.src_loc()!r}'
						)
					self.add_trigger(domain.clk, trigger = 1 if domain.clk_edge == 'pos' else 0)
					if domain.rst is not None and domain.async_reset:
						self.add_trigger(domain.rst, trigger = 1)
					return

				elif type(command) is Settle:
					self.state.wait_interval(self, None)
					return

				elif type(command) is Delay:
					# Internal timeline is in 1ps integral units, intervals are public API and in floating point
					interval = int(command.interval * 1e12) if command.interval is not None else None
					self.state.wait_interval(self, interval)
					return

				elif type(command) is Passive:
					self.passive = True

				elif type(command) is Active:
					self.passive = False

				elif command is None: # only possible if self.default_cmd is None
					raise TypeError(
						f'Received default command from process {self.src_loc()!r} that was added '
						'with add_process(); did you mean to add this process with '
						'add_sync_process() instead?'
					)

				else:
					raise TypeError(f'Received unsupported command {command!r} from process {self.src_loc()!r}')

			except Exception as exn:
				response = None
				exception = exn



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from contextlib           import contextmanager

from torii.hdl.ast        import Signal, Value, Statement
from torii.hdl.dsl        import Module
from torii.hdl.ir         import Fragment
from torii.sim            import Settle, Simulator
from torii.util           import flatten

from torii_sim_wasm       import WASMSimEngine

from ..utils              import ToriiTestSuiteCase
from .integration_harness import SimulatorIntegrationTestsMixin
from .regression_harness  import SimulatorRegressionTestMixin
from .unitest_harness     import SimulatorUnitTestsMixin

# TODO(aki): Figure out a better name
class SimulatorEngineTestCase(ToriiTestSuiteCase):
	def test_external_sim_engine(self):
		from torii.sim._base import BaseEngine

		class DummyEngine(BaseEngine):
			def __init__(self, fragment) -> None:
				pass

		_ = Simulator(Module(), engine = DummyEngine)

	def test_invalid_simulator_engine(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^The specified engine \'NotAValidEngineName\' is not a known simulation engine name, or simulation '
			'engine class$'
		):
			_ = Simulator(Module(), engine = 'NotAValidEngineName') # type: ignore

		with self.assertRaisesRegex(
			TypeError,
			r'^The specified engine <class \'object\'> is not a known simulation engine name, or simulation '
			'engine class$'
		):
			_ = Simulator(Module(), engine = object) # type: ignore

class WASMSimulatorUnitTestCase(ToriiTestSuiteCase, SimulatorUnitTestsMixin):
	def assertStatement(self, stmt, inputs, output, reset = 0):
		inputs = [Value.cast(i) for i in inputs]
		output = Value.cast(output)

		isigs = [ Signal(i.shape(), name = n) for i, n in zip(inputs, 'abcd') ]
		osig  = Signal(output.shape(), name = 'y', reset = reset)

		stmt = stmt(osig, *isigs)
		frag = Fragment()
		frag.add_statements(stmt)
		for signal in flatten(s._lhs_signals() for s in Statement.cast(stmt)):
			frag.add_driver(signal)

		sim = Simulator(frag, engine = WASMSimEngine)

		def process():
			for isig, input in zip(isigs, inputs):
				yield isig.eq(input)
			yield Settle()
			self.assertEqual((yield osig), output.value)

		sim.add_process(process)
		with sim.write_vcd('test.vcd', 'test.gtkw', traces = [ *isigs, osig ]):
			sim.run()

class WASMSimulatorIntegrationTestCase(ToriiTestSuiteCase, SimulatorIntegrationTestsMixin):
	@contextmanager
	def assertSimulation(self, module, deadline = None):
		sim = Simulator(module, engine = WASMSimEngine)
		yield sim
		with sim.write_vcd('test.vcd', 'test.gtkw'):
			if deadline is None:
				sim.run()
			else:
				sim.run_until(deadline)

class WASMRegressionTestCase(ToriiTestSuiteCase, SimulatorRegressionTestMixin):
	def get_simulator(self, dut) -> Simulator:
		return Simulator(dut, engine = WASMSimEngine)



# SPDX-License-Identifier: BSD-2-Clause

from torii.sim._base import BaseProcess

__all__ = (
	'WASMClockProcess',
)

class WASMClockProcess(BaseProcess):
	def __init__(self, state, signal, *, phase, period) -> None:
		if len(signal) != 1:
			raise TypeError(f'Clock signal must be exactly 1-wide, not {len(signal)}')

		self.state  = state
		self.slot   = self.state.get_signal(signal)
		self.phase  = phase
		self.period = period

		self.reset()

	def reset(self):
		self.runnable = True
		self.passive = True

		self.initial = True

	def run(self):
		self.runnable = False

		if self.initial:
			self.initial = False
			self.state.wait_interval(self, self.phase)

		else:
			clk_state = self.state.slots[self.slot]
			clk_state.update(not clk_state.curr.value())
			self.state.wait_interval(self, self.period // 2)



# SPDX-License-Identifier: BSD-2-Clause

from os           import getenv
from pathlib      import Path

import nox
from nox.sessions import Session

ROOT_DIR  = Path(__file__).parent

BUILD_DIR = ROOT_DIR  / 'build'
CNTRB_DIR = ROOT_DIR  / 'contrib'
DOCS_DIR  = ROOT_DIR  / 'docs'
DIST_DIR  = BUILD_DIR / 'dist'

IN_CI           = getenv('GITHUB_WORKSPACE') is not None
ENABLE_COVERAGE = IN_CI or getenv('TORII_TEST_COVERAGE') is not None

# Default sessions to run
nox.options.sessions = (
	'test',
	'lint',
	'typecheck-mypy'
)

# Try to use `uv`, if not fallback to `virtualenv`
nox.options.default_venv_backend = 'uv|virtualenv'

@nox.session(reuse_venv = True)
def test(session: Session) -> None:
	OUTPUT_DIR = BUILD_DIR / 'tests'
	OUTPUT_DIR.mkdir(parents = True, exist_ok = True)

	unittest_args = ('-m', 'unittest', 'discover', '-s', str(ROOT_DIR / 'python'))

	session.install('-e', '.')

	if ENABLE_COVERAGE:
		session.log('Coverage support enabled')
		session.install('coverage')
		coverage_args = ('-m', 'coverage', 'run', '-p', f'--rcfile={ROOT_DIR / "pyproject.toml"}',)
		session.env['COVERAGE_CORE'] = 'sysmon'
	else:
		coverage_args = tuple[str]()

	with session.chdir(OUTPUT_DIR):
		session.log('Running core test suite...')
		session.run('python', *coverage_args, *unittest_args, *session.posargs)

		if ENABLE_COVERAGE:
			session.log('Combining Coverage data..')
			session.run('python', '-m', 'coverage', 'combine')

			session.log('Generating XML Coverage report...')
			session.run('python', '-m', 'coverage', 'xml', f'--rcfile={ROOT_DIR / "pyproject.toml"}')

# TODO(aki): Figure out how the documentation will be handled. (Part of Torii itself?)

@nox.session(name = 'typecheck-mypy', reuse_venv = True)
def typecheck_mypy(session: Session) -> None:
	OUTPUT_DIR = BUILD_DIR / 'typing' / 'mypy'
	OUTPUT_DIR.mkdir(parents = True, exist_ok = True)

	session.install('mypy')
	session.install('lxml')
	session.install('-e', '.')

	session.run(
		'mypy', '--non-interactive', '--install-types', '--pretty',
		'--disallow-any-generics',
		'--cache-dir', str((OUTPUT_DIR / '.mypy-cache').resolve()),
		'-p', 'torii_sim_wasm', '--html-report', str(OUTPUT_DIR.resolve())
	)

@nox.session(name = 'typecheck-pyright', reuse_venv = True)
def typecheck_pyright(session: Session) -> None:
	OUTPUT_DIR = BUILD_DIR / 'typing' / 'pyright'
	OUTPUT_DIR.mkdir(parents = True, exist_ok = True)

	session.install('pyright')
	session.install('types-Pygments', 'types-setuptools')
	session.install('-e', '.')

	with (OUTPUT_DIR / 'pyright.log').open('w') as f:
		session.run('pyright', *session.posargs, stdout = f)

@nox.session(reuse_venv = True)
def lint(session: Session) -> None:
	session.install('flake8')
	session.install('ruff')

	session.run(
		'flake8', '--config', str((CNTRB_DIR / '.flake8').resolve()),
		'./python/torii_sim_wasm', './python/tests',
	)
	session.run('ruff', 'check', './python/torii_sim_wasm', './python/tests')

@nox.session(reuse_venv = True)
def dist(session: Session) -> None:
	session.install('build')

	session.run('python', '-m', 'build', '-o', str(DIST_DIR))



# SPDX-License-Identifier: BSD-2-Clause

from contextlib  import contextmanager
from os          import getenv
from tempfile    import NamedTemporaryFile

from torii.hdl.ast   import SignalSet
from torii.hdl.ir    import Fragment
from torii.hdl.xfrm  import LHSGroupFilter, StatementVisitor, ValueVisitor
from torii.sim._base import BaseProcess

from ._wasm_engine   import WASMRunner

__all__ = (
	'WASMFragmentCompiler',
	'WASMRTLProcess',
)

WASM_SET_SLOT = '''
	(func $slots_set (param $index i64) (param $value i64)
		(local $curr i64)
		(local $next i64)
		(local $next_off i64)
		(local.set $curr (i64.load (i64.mul (local.get $index) (i64.const 16))))
		(local.set $next_off (i64.mul (i64.add (i64.mul (local.get $index) (i64.const 2)) (i64.const 1)) (i64.const 8)))
		(local.set $next (i64.load (local.get $next_off)))
		(if (i64.ne (local.get $next) (local.get $value))
			(then
				(i64.store (local.get $next_off) (local.get $value))
				(call $slots_set_py (local.get $index) (local.get $value))
			)
		)
	)
'''

WASM_SIGN = '''
	(func $sign (param $value i64) (param $sign i64) (result i64)
		(if (result i64) (i64.ne (i64.and (local.get $value) (local.get $sign)) (i64.const 0))
			(then (return (i64.or (local.get $value) (local.get $sign))))
			(else (return (local.get $value)))
		)
	)
'''

# Signed floor div for integers
WASM_ZDIV = '''
	(func $zdiv (param $lhs i64) (param $rhs i64) (result i64)
		(local $res i64)
		(if (result i64) (i64.eq (local.get $rhs) (i64.const 0))
			(then (return (i64.const 0)))
			(else
				(local.set $res (i64.div_s (local.get $lhs) (local.get $rhs)))
				(if (i32.gt_u (i32.and
							(i64.lt_s (i64.xor (local.get $lhs) (local.get $rhs)) (i64.const 0))
							(i64.ne (i64.rem_s (local.get $lhs) (local.get $rhs)) (i64.const 0))
						)
						(i32.const 0)
					)
					(then (local.set $res (i64.sub (local.get $res) (i64.const 1))))
				)
				(return (local.get $res))
			)
		)
	)
'''

WASM_ZMOD = '''
	(func $zmod (param $lhs i64) (param $rhs i64) (result i64)
		(if (result i64) (i64.eq (local.get $rhs) (i64.const 0))
			(then (return (i64.const 0)))
			(else (return
				(i64.rem_s
					(i64.add
						(i64.rem_s (local.get $lhs) (local.get $rhs))
						(local.get $rhs)
					)
					(local.get $rhs)
				)
			))
		)
	)
'''

class WASMRTLProcess(BaseProcess):
	__slots__ = ('is_comb', 'runnable', 'passive', 'run')

	def __init__(self, *, is_comb) -> None:
		self.is_comb  = is_comb

		self.reset()

	def reset(self):
		self.runnable = self.is_comb
		self.passive  = True

class _WASMEmitter:
	def __init__(self):
		self._level = 0
		self._suffix = 0
		self._imports = []
		self._globals = []
		self._variables = []
		self._instructions = []

	def add_src(self, src_loc):
		if src_loc:
			self.append(';; {}:{}'.format(*src_loc))

	def append(self, code):
		self._instructions.append('\t\t' + '\t' * self._level)
		self._instructions.append(code)
		self._instructions.append('\n')

	def add_variable(self, name):
		self._variables.append('\t\t')
		self._variables.append(f'(local ${name} i64)')
		self._variables.append('\n')

	def def_var(self, name, code):
		name = f'{name}_{self._suffix}'
		self._suffix += 1
		self.add_variable(name)
		self.append(f'(local.set ${name} {code})')
		return name

	@contextmanager
	def indent(self):
		self._level += 1
		yield
		self._level -= 1

	def flush(self, result: bool = False):
		module = '(module\n'
		module += '\t(import "" "gmem" (memory $gmem i64 0 2 shared ))\n'
		module += '\t(func $slots_set_py (import "" "slots_set_py") (param i64) (param i64))\n'
		module += ''.join(self._globals)
		module += '\n'
		module += ''.join(self._imports)
		module += '\n'
		module += WASM_SET_SLOT
		module += '\n'
		module += WASM_SIGN
		module += '\n'
		module += WASM_ZDIV
		module += '\n'
		module += WASM_ZMOD
		module += '\n'
		module += '\t(func (export "run") (result i64)\n'
		module += ''.join(self._variables)
		module += ''.join(self._instructions)
		if not result:
			module += '\t\t(i64.const 0)\n'
		module += "\n\t)\n)\n"

		self._instructions.clear()
		return module

class _Compiler:
	def __init__(self, state, emitter) -> None:
		self.state = state
		self.emitter = emitter

class _ValueCompiler(ValueVisitor, _Compiler):
	def on_value(self, value):
		# Wasm currently only supports 63 bit wide values
		if len(value) > 63:
			if value.src_loc:
				src = '{}:{}'.format(*value.src_loc)
			else:
				src = 'unknown location'
			raise OverflowError(
				f'Value defined at {src} is {len(value)} bits wide, and wasm backend only supports '
				'signals that are less than 64 bits wide'
			)

		val = super().on_value(value)
		return val

	def on_ClockSignal(self, value):
		raise NotImplementedError # :nocov:

	def on_ResetSignal(self, value):
		raise NotImplementedError # :nocov:

	def on_AnyValue(self, value):
		raise NotImplementedError # :nocov:

	def on_Sample(self, value):
		raise NotImplementedError # :nocov:

	def on_Initial(self, value):
		raise NotImplementedError # :nocov:

class _RHSValueCompiler(_ValueCompiler):
	def __init__(self, state, emitter, *, mode, inputs = None) -> None:
		super().__init__(state, emitter)
		if mode not in ('curr', 'next'):
			raise ValueError(f'Expected mode to be \'curr\', or \'next\', not \'{mode!r}\'')
		self.mode = mode
		# If not None, `inputs` gets populated with RHS signals.
		self.inputs = inputs

	def on_Const(self, value):
		return f'(i64.const {value.value})'

	def on_Signal(self, value):
		if self.inputs is not None:
			self.inputs.add(value)

		if self.mode == 'curr':
			return f'(i64.load (i64.const {self.state.get_signal(value) * 16}))'
		else:
			return f'(local.get $next_{self.state.get_signal(value)})'

	def on_Operator(self, value):
		def mask(value):
			value_mask = (1 << len(value)) - 1
			return f'(i64.and (i64.const {value_mask:#x}) {self(value)})'

		def sign(value):
			if value.shape().signed:
				return f'(call $sign {mask(value)} (i64.const {-1 << (len(value) - 1):#x}))'
			else: # unsigned
				return mask(value)
			return mask(value)

		if len(value.operands) == 1:
			arg, = value.operands
			if value.operator == '~':
				return f'(i64.xor {mask(arg)} (i64.const 0xffffffffffffffff))'
			if value.operator == '-':
				return f'(i64.mul {sign(arg)} (i64.const -1))'
			if value.operator == 'b':
				return f'(i64.extend_i32_u (i64.gt_u {mask(arg)} (i64.const 0)))'
			if value.operator == 'r|':
				return f'(i64.extend_i32_u (i64.ne {mask(arg)} (i64.const 0)))'
			if value.operator == 'r&':
				return f'(i64.extend_i32_u (i64.eq {mask(arg)} (i64.const {(1 << len(arg)) - 1})))'
			if value.operator == 'r^':
				return f'(i64.rem_u (i64.popcnt {mask(arg)}) (i64.const 2))'
			if value.operator in ('u', 's'):
				# These operators don't change the bit pattern, only its interpretation.
				return self(arg)
		elif len(value.operands) == 2:
			lhs, rhs = value.operands
			if value.operator == '+':
				return f'(i64.add {sign(lhs)} {sign(rhs)})'
			if value.operator == '-':
				return f'(i64.sub {sign(lhs)} {sign(rhs)})'
			if value.operator == '*':
				return f'(i64.mul {sign(lhs)} {sign(rhs)})'
			if value.operator == '//':
				return f'(call $zdiv {sign(lhs)} {sign(rhs)})'
			if value.operator == '%':
				return f'(call $zmod {sign(lhs)} {sign(rhs)})'
			if value.operator == '&':
				return f'(i64.and {sign(lhs)} {sign(rhs)})'
			if value.operator == '|':
				return f'(i64.or {sign(lhs)} {sign(rhs)})'
			if value.operator == '^':
				return f'(i64.xor {sign(lhs)} {sign(rhs)})'
			if value.operator == '<<':
				return f'(i64.shl {sign(lhs)} {sign(rhs)})'
			if value.operator == '>>':
				return f'(i64.shr_u {sign(lhs)} {sign(rhs)})'
			if value.operator == '!=':
				# i64.eq will push i32 into a stack so we need to extend it to i64
				return f'(i64.extend_i32_u (i64.ne {sign(lhs)} {sign(rhs)}))'
			if value.operator == '<':
				return f'(i64.extend_i32_u (i64.lt_s {sign(lhs)} {sign(rhs)}))'
			if value.operator == '<=':
				return f'(i64.extend_i32_u (i64.le_s {sign(lhs)} {sign(rhs)}))'
			if value.operator == '>':
				return f'(i64.extend_i32_u (i64.gt_s {sign(lhs)} {sign(rhs)}))'
			if value.operator == '>=':
				return f'(i64.extend_i32_u (i64.ge_s {sign(lhs)} {sign(rhs)}))'
			if value.operator == '==':
				return f'(i64.extend_i32_u (i64.eq {sign(lhs)} {sign(rhs)}))'
		elif len(value.operands) == 3:
			if value.operator == 'm':
				sel, val1, val0 = value.operands
				return f'(if (result i64) (i64.gt_u {mask(sel)} (i64.const 0)) (then {sign(val1)}) (else {sign(val0)}))'
		raise NotImplementedError(f'Operator \'{value.operator}\' not implemented') # :nocov:

	def on_Slice(self, value):
		and_shift = f'(i64.shr_u {self(value.value)} (i64.const {value.start}))'
		return f'(i64.and (i64.const {(1 << len(value)) - 1:#x}) {and_shift})'

	def on_Part(self, value):
		offset_mask = (1 << len(value.offset)) - 1
		offset = f'(i64.mul (i64.const {value.stride}) (i64.and (i64.const {offset_mask:#x}) {self(value.offset)}))'
		return f'(i64.and (i64.const {(1 << value.width) - 1}) (i64.shr_u {self(value.value)} {offset}))'

	def on_Cat(self, value):
		gen_parts = []
		offset = 0
		for part in value.parts:
			part_mask = (1 << len(part)) - 1
			gen_parts.append(f'(i64.shl (i64.and (i64.const {part_mask:#x}) {self(part)}) (i64.const {offset}))')
			offset += len(part)

		# we have to nest the or statements so time to do annoying paren stuff
		if gen_parts:
			return f'{"(i64.or ".join(gen_parts)}{")" * (len(gen_parts) - 1)}'
		return '(i64.const 0)'

	def on_ArrayProxy(self, value):
		index_mask = (1 << len(value.index)) - 1
		self.emitter.add_src(value.src_loc)
		gen_index = self.emitter.def_var('rhs_index', f'(i64.and (i64.const {index_mask:#x}) {self(value.index)})')
		gen_value = self.emitter.def_var('rhs_proxy', '(i64.const 0)')
		if value.elems:
			for index, elem in enumerate(value.elems):
				self.emitter.add_src(value.src_loc)
				check = f'(i64.eq (i64.const {index}) (local.get ${gen_index}))'
				if index == 0:
					self.emitter.append(f'(if {check} (then')
				else:
					self.emitter.append(f'(else (if {check} (then')
				with self.emitter.indent():
					self.emitter.append(f'(local.set ${gen_value} {self(elem)})')
				self.emitter.append(')')

			self.emitter.append('(else')
			with self.emitter.indent():
				self.emitter.append(f'(local.set ${gen_value} {self(value.elems[-1])})')

			self.emitter.append('))' + '))' * (len(value.elems) - 1))
			return f'(local.get ${gen_value})'
		else:
			return '(i64.const 0)'

	@classmethod
	def compile(cls, state, value, *, mode):
		emitter = _WASMEmitter()
		compiler = cls(state, emitter, mode = mode)
		emitter.append(compiler(value))

		output_code = emitter.flush(True)
		return output_code

class _LHSValueCompiler(_ValueCompiler):
	def __init__(self, state, emitter, *, rhs, outputs = None) -> None:
		super().__init__(state, emitter)
		# `rrhs` is used to translate rvalues that are syntactically a part of an lvalue, e.g.
		# the offset of a Part.
		self.rrhs = rhs
		# `lrhs` is used to translate the read part of a read-modify-write cycle during partial
		# update of an lvalue.
		self.lrhs = _RHSValueCompiler(state, emitter, mode = 'next', inputs = None)
		# If not None, `outputs` gets populated with signals on LHS.
		self.outputs = outputs

	def on_Const(self, value):
		raise TypeError # :nocov:

	def on_Signal(self, value):
		if self.outputs is not None:
			self.outputs.add(value)

		def gen(arg):
			value_mask = (1 << len(value)) - 1
			if value.shape().signed:
				value_const = f'(i64.const {-1 << (len(value) - 1):#x})'
				value_sign = f'(call $sign (i64.and (i64.const {value_mask:#x}) {arg}) {value_const})'
			else: # unsigned
				value_sign = f'(i64.and (i64.const {value_mask:#x}) {arg})'
			self.emitter.add_src(value.src_loc)
			self.emitter.append(f'(local.set $next_{self.state.get_signal(value)} {value_sign})')
		return gen

	def on_Operator(self, value):
		if value.operator in ('u', 's'):
			return self(value.operands[0])
		raise TypeError # :nocov:

	def on_Slice(self, value):
		def gen(arg):
			width_mask = (1 << (value.stop - value.start)) - 1
			self(value.value)(
				f'(i64.or '
				f'(i64.and {self.lrhs(value.value)} '
				f'(i64.const {~(width_mask << value.start):#x})) '
				f'(i64.shl (i64.and (i64.const {width_mask:#x}) {arg}) (i64.const {value.start}))'
				f')'
			)
		return gen

	def on_Part(self, value):
		def gen(arg):
			width_mask = (1 << value.width) - 1
			offset_mask = (1 << len(value.offset)) - 1
			offset_and = f'(i64.and (i64.const {offset_mask:#x}) {self.rrhs(value.offset)})'
			offset = f'(i64.mul (i64.const {value.stride}) {offset_and})'
			self(value.value)(
				f'(i64.and {self.lrhs(value.value)} '
				f'(i64.or '
				f'(i64.xor (i64.shl (i64.const {width_mask:#x}) {offset}) (i64.const 0xffffffffffffffff)) '
				f'(i64.shl (i64.and (i64.const {width_mask:#x}) {arg}) {offset}))'
				f')'
			)
		return gen

	def on_Cat(self, value):
		def gen(arg):
			self.emitter.add_src(value.src_loc)
			gen_arg = self.emitter.def_var('cat', arg)
			offset = 0
			for part in value.parts:
				part_mask = (1 << len(part)) - 1
				part_shift = f'(i64.shr_u (local.get ${gen_arg}) (i64.const {offset}))'
				self(part)(f'(i64.and (i64.const {part_mask:#x}) {part_shift})')
				offset += len(part)
		return gen

	def on_ArrayProxy(self, value):
		def gen(arg):
			self.emitter.add_src(value.src_loc)
			index_mask = (1 << len(value.index)) - 1
			gen_index = self.emitter.def_var('index', f'(i64.and {self.rrhs(value.index)} (i64.const {index_mask:#x}))')
			if value.elems:
				for index, elem in enumerate(value.elems):
					self.emitter.add_src(value.src_loc)
					check = f'(i64.eq (i64.const {index}) (local.get ${gen_index}))'
					if index == 0:
						self.emitter.append(f'(if {check} (then')
					else:
						self.emitter.append(f'(else (if {check} (then')
					with self.emitter.indent():
						self(elem)(arg)
					self.emitter.append(')')

				self.emitter.add_src(value.src_loc)
				self.emitter.append('(else')
				with self.emitter.indent():
					self(value.elems[-1])(arg)

				self.emitter.append('))' + '))' * (len(value.elems) - 1))
		return gen

class _StatementCompiler(StatementVisitor, _Compiler):
	def __init__(self, state, emitter, *, inputs = None, outputs = None) -> None:
		super().__init__(state, emitter)
		self.rhs = _RHSValueCompiler(state, emitter, mode = 'curr', inputs = inputs)
		self.lhs = _LHSValueCompiler(state, emitter, rhs = self.rhs, outputs = outputs)

	def on_statements(self, stmts):
		for stmt in stmts:
			self(stmt)

	def on_Assign(self, stmt):
		gen_rhs_value = self.rhs(stmt.rhs) # check for oversized value before generating mask
		gen_rhs = f'(i64.and (i64.const {(1 << len(stmt.rhs)) - 1:#x}) {gen_rhs_value})'
		if stmt.rhs.shape().signed:
			gen_rhs = f'(call $sign {gen_rhs} (i64.const {-1 << (len(stmt.rhs) - 1):#x}))'
		return self.lhs(stmt.lhs)(gen_rhs)

	def on_Switch(self, stmt):
		test_value = self.rhs(stmt.test) # check for oversized value before generating mask
		gen_test = self.emitter.def_var('test', f'(i64.and (i64.const {(1 << len(stmt.test)) - 1:#x}) {test_value})')

		for index, (patterns, stmts) in enumerate(stmt.cases.items()):
			self.emitter.add_src(stmt.src_loc)
			gen_checks = []
			if not patterns:
				gen_checks.append('(i32.eq (i32.const 1) (i32.const 1))')
			else:
				for pattern in patterns:
					if "-" in pattern:
						mask  = int(''.join('0' if b == '-' else '1' for b in pattern), 2)
						value = int(''.join('0' if b == '-' else b for b in pattern), 2)
						value_and = f'(i64.and (i64.const {mask}) (local.get ${gen_test}))'
						gen_checks.append(f'(i64.eq (i64.const {value}) {value_and})')
					else:
						value = int(pattern or '0', 2)
						gen_checks.append(f'(i64.eq (i64.const {value}) (local.get ${gen_test}))')

			if index == 0:
				self.emitter.append(f'(if {"".join(gen_checks)} (then')
			else:
				self.emitter.append(f'(else (if {"".join(gen_checks)} (then')

			with self.emitter.indent():
				self(stmts)

			self.emitter.append(')')

		# Close down all the nested if-elses
		if len(stmt.cases.items()) > 0:
			self.emitter.append(')' + '))' * (len(stmt.cases.items()) - 1))

	def on_Property(self, stmt):
		raise NotImplementedError # :nocov:

	@classmethod
	def compile(cls, state, stmt):
		output_indexes = [state.get_signal(signal) for signal in stmt._lhs_signals()]
		emitter = _WASMEmitter()
		for signal_index in output_indexes:
			emitter.add_variable(f'next_{signal_index}')
			emitter.append(f'(local.set $next_{signal_index} (i64.load (i64.const {(signal_index * 2 + 1) * 8})))')
		compiler = cls(state, emitter)
		compiler(stmt)
		for signal_index in output_indexes:
			emitter.append(f'(call $slots_set (i64.const {signal_index}) (local.get $next_{signal_index}))')

		output_code = emitter.flush()
		return output_code

class WASMFragmentCompiler:
	def __init__(self, state) -> None:
		self.state = state

	def __call__(self, fragment: Fragment):
		processes = set()

		for domain_name, domain_signals in fragment.drivers.items():
			domain_stmts = LHSGroupFilter(domain_signals)(fragment.statements)
			domain_process = WASMRTLProcess(is_comb = domain_name is None)

			emitter = _WASMEmitter()
			if domain_name is None:
				for signal in domain_signals:
					signal_index = self.state.get_signal(signal)
					emitter.add_variable(f'next_{signal_index}')
					emitter.append(f'(local.set $next_{signal_index} (i64.const {signal.reset}))')

				inputs = SignalSet()
				_StatementCompiler(self.state, emitter, inputs = inputs)(domain_stmts)

				for input in inputs:
					self.state.add_trigger(domain_process, input)
			else:
				domain = fragment.domains[domain_name]
				clk_trigger = 1 if domain.clk_edge == 'pos' else 0
				self.state.add_trigger(domain_process, domain.clk, trigger = clk_trigger)
				if domain.rst is not None and domain.async_reset:
					rst_trigger = 1
					self.state.add_trigger(domain_process, domain.rst, trigger = rst_trigger)

				for signal in domain_signals:
					signal_index = self.state.get_signal(signal)
					emitter.add_variable(f'next_{signal_index}')
					index_const = f'(i64.const {(signal_index * 2 + 1) * 8})'
					emitter.append(f'(local.set $next_{signal_index} (i64.load {index_const}))')

				_StatementCompiler(self.state, emitter)(domain_stmts)

			for signal in domain_signals:
				signal_index = self.state.get_signal(signal)
				emitter.append(f'(call $slots_set (i64.const {signal_index}) (local.get $next_{signal_index}))')

			module_code = emitter.flush()
			if getenv('TORII_WASMSIM_DUMP'):
				file = NamedTemporaryFile('w', prefix = 'torii_wasmsim_', delete = False)
				file.write(module_code)

			domain_process.run = WASMRunner(module_code, self.state.memory, self.state.set_slot)
			processes.add(domain_process)

		for subfragment_index, (subfragment, subfragment_name) in enumerate(fragment.subfragments):
			if subfragment_name is None:
				subfragment_name = f'U${subfragment_index}'
			processes.update(self(subfragment))

		return processes



# SPDX-License-Identifier: BSD-2-Clause

import os
import re
import shutil
import subprocess
import textwrap
import traceback
from pathlib       import Path

from torii.back    import rtlil
from torii.hdl.ast import Statement
from torii.hdl.ir  import Fragment
from torii.test    import ToriiTestCase
from torii.tools   import ToolNotFound, require_tool

__all__ = (
	'ToriiTestSuiteCase',
)

class ToriiTestSuiteCase(ToriiTestCase):
	def assertRepr(self, obj, repr_str):
		if isinstance(obj, list):
			obj = Statement.cast(obj)

		def prepare_repr(repr_str):
			repr_str = re.sub(r'\s+', ' ',  repr_str)
			repr_str = re.sub(r'\( (?=\()', '(', repr_str)
			repr_str = re.sub(r'\) (?=\))', ')', repr_str)
			return repr_str.strip()
		self.assertEqual(prepare_repr(repr(obj)), prepare_repr(repr_str))

	# TODO: Once the Torii formal bits are better defined remove this and add formal to ToriiTestCase
	def assertFormal(self, spec, mode = 'bmc', depth = 1):
		stack      = traceback.extract_stack()
		file_dir   = Path(__file__).resolve().parent
		formal_dir = Path.cwd() / 'torii-formal'

		if not formal_dir.exists():
			formal_dir.mkdir(parents = True, exist_ok = True)

		for frame in reversed(stack):
			if str(file_dir) not in frame.filename:
				break
			caller = frame

		fname = Path(caller.filename).resolve().stem.replace('test_', 'spec_')
		spec_name = f'{fname}_{caller.name.replace("test_", "")}'
		spec_dir = formal_dir / spec_name

		# The sby -f switch seems not fully functional when sby is reading from stdin.
		if spec_dir.exists():
			shutil.rmtree(spec_dir)

		if mode == 'hybrid':
			# A mix of BMC and k-induction, as per personal communication with Claire Wolf.
			script = 'setattr -unset init w:* a:torii.sample_reg %d'
			mode   = 'bmc'
		else:
			script = ''

		config = textwrap.dedent(f'''\
		[options]
		mode {mode}
		depth {depth}
		wait on
		multiclock on

		[engines]
		smtbmc

		[script]
		read_rtlil top.il
		prep
		{script}

		[file top.il]
		{rtlil.convert_fragment(Fragment.get(spec, platform = 'formal').prepare())[0]}
		''')

		try:
			sby = require_tool('sby')
		except ToolNotFound: # :nocov:
			self.skipTest('SBY not installed')

		# We don't actually use click, but SBY does so if it's missing it'll blow up
		try:
			import click
		except ImportError: # :nocov:
			self.skipTest('SBY is installed but click is not, SBY won\'t run')

		del click

		with subprocess.Popen([
				sby, '-f', '-d', spec_name
			],
			cwd                = formal_dir,
			env                = {
				**os.environ,
				'PYTHONWARNINGS': 'ignore'
			},
			universal_newlines = True,
			stdin              = subprocess.PIPE,
			stdout             = subprocess.PIPE
		) as proc:
			stdout, stderr = proc.communicate(config)
			if proc.returncode != 0:
				self.fail('Formal verification failed:\n' + stdout)
